/***************************************************************************
 *   Copyright (C) 2008 by Emil Ljungdahl <emil@kratern.se>                *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA .        *
 *                                                                         *
 *   $Id: quicklauncher.h 14 2008-08-20 20:48:47Z emil $                    *
 ***************************************************************************/

#ifndef QUICKLAUNCHER_H
#define QUICKLAUNCHER_H

// QT
#include <QGraphicsSceneContextMenuEvent>
#include <QGraphicsSceneDragDropEvent>
#include <QGraphicsSceneMouseEvent>
#include <QGraphicsView>
#include <QMimeData>
#include <QDrag>
#include <QGraphicsGridLayout>
#include <QGraphicsLinearLayout>

// KDE
#include <KUrl>
#include <KRun>
#include <KDialog>
#include <KMimeType>
#include <KGlobalSettings>
#include <KSharedConfig>
#include <KDesktopFile>

// Plasma
#include <plasma/widgets/icon.h>
#include <Plasma/Applet>
#include <Plasma/Containment>

#include <math.h>

#include "ui_quicklauncherConfig.h"
#include "ui_quicklauncherAdd.h"

/**
 * Extends Plasma::Icon with execution of url
 */
class MyIcon : public Plasma::Icon
{
	Q_OBJECT
	public:
		/**
		 * Constructor
		 * @param mimetype String with mimetype, used for exec desitions
		 * @param appUrl Url to the .desktop-file (or whatever)
		 * @param parent Parent item
		 */
		MyIcon(QString mimetype, KUrl appUrl, QGraphicsItem *parent = 0) : Plasma::Icon(parent), mimetype_(mimetype), appUrl_(appUrl), hidden_(false)
		{
			setWindowFrameMargins(0,0,0,0);
			connect(this, SIGNAL(clicked()), this, SLOT(openUrl()));
		}

		/**
		 * The url for this icon
		 * @return Url as KUrl object
		 */
		KUrl getUrl() { return appUrl_; }

		/**
		 * The url for this icon in string format
		 * @return Url (or path) for this icon
		 */
		QString getPath() { return appUrl_.path(); }

		/**
		 * Whether the icon should be hidden or not
		 * @return true if icon is hidden, else false
		 */
		bool isHidden() { return hidden_; }

		/**
		 * Tell whether the icon should be hidden or not
		 * @param hidden Set to true to hide icon
		 */
		void setHidden(bool hidden) { hidden_ = hidden; }

	private slots:
		/**
		 * Slot for open the url when the icon is clicked
		 */
		void openUrl()
		{
			KRun::runUrl(appUrl_, mimetype_, 0);
		}
	private:
		/** The mimetype for the Url */
		QString mimetype_;
		/** Url to file */
		KUrl appUrl_;
		/** Should the icon be hidden? */
		bool hidden_;
};

class InnerLayout : public QGraphicsGridLayout
{
	public:
		InnerLayout(QGraphicsLayoutItem *parent, int rowCount)
			: QGraphicsGridLayout(parent), m_rowCount(rowCount)
		{
			setVerticalSpacing(0);
			setHorizontalSpacing(0);
		}
		void setRowCount(int rowCount) { m_rowCount = rowCount; }
		void addItem(Plasma::Icon *icon) {
			int row = (rowCount() == -1) ? 0 : (count() % m_rowCount);
			int column = columnCount() - ((row > 0) ? 1 : 0);
			QGraphicsGridLayout::addItem(icon, row, column);
		}

		QRectF geometry() const
		{
			qreal minRowSize = 1;
			qreal minColSize = 1;
			for(int i = 0; i < count(); i++) {
				minRowSize = qMax(minRowSize, itemAt(i)->geometry().height());
				minColSize = qMax(minColSize, itemAt(i)->geometry().width());
			}
			int colCount = ((count() - 1) / m_rowCount) + 1;
			return QRectF(0,0,minColSize*colCount,minRowSize*m_rowCount);
		}
	private:
		int m_rowCount;
};

/**
 * QuickLauncher applet
 */
class QuickLauncher : public Plasma::Applet
{
	Q_OBJECT

	public:
		/**
		 * Constructor
		 */
		QuickLauncher(QObject *parent, const QVariantList &args);
		/**
		 * Destructor
		 */
		~QuickLauncher();
		/**
		 * Initialization method
		 */
		void init();

		/**
		 * Returns hints about the geometry of the figure
		 * @return Hints about proportionality of the applet
		 */
		QSizeF contentSizeHint() const;

		/**
		 * List of actions to add in context menu
		 * @return List of QAction pointers
		 */
		virtual QList<QAction*> contextualActions();

	public slots:
		/**
		 * Overloaded method to show configuration interface
		 */
		void showConfigurationInterface();

		/**
		 * Slot for showing the Add Icon interface
		 */
		void showAddInterface();

		/**
		 * Slot to remove an icon
		 * Which icon to remove is determined by internal mouse event listeners
		 */
		void removeIcon();

		/**
		 * Slot to toggle the visibility of hidden icons
		 */
		void toggleShowAllIcons();

		/**
		 * Hides icons which are marked as hidden
		 */
		void hideIcons();

	protected slots:
		/**
		 * Called when the user has clicked OK in the Add Icon interface
		 */
		void addAccepted();

		/**
		 * Called when the user has clicked OK or apply in the Config interface
		 */
		void configAccepted();


		void constraintsEvent(Plasma::Constraints constraints);

	protected:
		/**
		 * Overloaded method to catch right-click events
		 * Used to save which icon we have clicked on (if any)
		 */
		void contextMenuEvent(QGraphicsSceneContextMenuEvent* event);

		/**
		 * Event filters to steel some events from our childs (icons)
		 */
		bool sceneEventFilter (QGraphicsItem*, QEvent *event);

		/**
		 * Overloaded mouse press event listener
		 * Saves the click-position to be able to recognize drags
		 * later on
		 */
		void mousePressEvent(QGraphicsSceneMouseEvent *event);

		/**
		 * Overloaded mouse move event listener
		 * Checks for a saved click-position and determine if we
		 * should start a drag
		 */
		void mouseMoveEvent(QGraphicsSceneMouseEvent *event);

		/**
		 * Overloaded drag enter event listener
		 * Check if the dragged item is valid
		 */
		void dragEnterEvent(QGraphicsSceneDragDropEvent *event);

		/**
		 * Overloaded drag move event listener
		 * Listens for drag moves
		 */
		void dragMoveEvent(QGraphicsSceneDragDropEvent* event);

		/**
		 * Overloaded drop event listener
		 * Determines if a dropped item is valid Url, and if so
		 * add the icon to the applet
		 */
		void dropEvent(QGraphicsSceneDragDropEvent* event);

	private:
		/**
		 * Initialize a new layout
		 */
		void createLayout();

		/**
		 * Fill a layout with icons
		 */
		void fillWidget();

		void updateLayout();

		/**
		 * Read a Url, and insert at the given position
		 * @param index The position to insert the icon into
		 * @param desktopFile The Url to read
		 */
		void readDesktopFile(int index, QString desktopFile);

		/**
		 * Read all Urls from a list, and insert into icon list
		 * @param desktopFiles List with Urls
		 */
		void readDestopFiles(QStringList desktopFiles);

		/**
		 * Remove icon at the given position
		 * @param iconIndex Index of icon to remove
		 */
		void removeIcon(int iconIndex);

		/**
		 * Finds an icon (if any) at the given position
		 * @param point Position where to look for an icon
		 * @return Index of a found icon, or -1 if none
		 */
		int findClickedIcon(QPointF point);

		/**
		 * Initialize drag operation
		 */
		void startDrag(QGraphicsSceneMouseEvent *event);

		/**
		 * Saves icons into plasma applet config file
		 */
		void saveConfig();

		KDialog *add_dialog;
		KDialog *conf_dialog;
		Ui::quicklauncherConfig uiConfig;
		Ui::quicklauncherAdd uiAdd;
		QList<MyIcon*> iconList;
		InnerLayout *innerLayout;
		Plasma::Icon *arrowIcon;
		int rowCount;
		int visibleIcons;
		QAction* addAction;
		QAction* removeAction;
		int rightClickedIcon;
		bool showAllIcons;

		QPointF startPos;
};

K_EXPORT_PLASMA_APPLET(quicklauncher, QuickLauncher)

#endif
